"""!
    The Python Ionic SDK handles errors through exceptions rather than error return codes.

Exception Classes:\n
    IonicException - Thrown when an error results from the client logic of the SDK.\n
    IonicServerException - Thrown as a result of an Ionic server
                           error and contains extra information such as
                           the HTTP response code and internal server
                           error code.\n
"""

import ionicsdk._util as _private
from ctypes import *

class CServerResponse(Structure):
    __slots__ = [
            'nHttpResponseCode',
            'nServerErrorCode',
            'pszServerErrorMessage',
            'pszServerErrorDataJson',
            'pszConversationId',
            ]
    _fields_ = [
            ('nHttpResponseCode', c_int),
            ('nServerErrorCode', c_int),
            ('pszServerErrorMessage', c_char_p),
            ('pszServerErrorDataJson', c_char_p),
            ('pszConversationId', c_char_p),
            ]

class IonicException(Exception):
    """!
        Exception for Ionic SDK client logic.
    
    Instance Variables:
        code (int): The error code that caused the exception.
                    See errors.py for more information.
    """
    def __init__(self, message, code):
        super(IonicException, self).__init__(message)
        ##(int) The error code that caused the exception. See errors.py for more information.
        self.code = code
        ##(int) The error text message
        self.message = message
    
    def __str__(self):
        message = super(IonicException, self).__str__()
        message += " - Error code: " + str(self.code)
        return message

class ServerResponse(object):
    """!
        Server response object returned from most Ionic server calls.
        
    Instance Variables:\n
        httpcode (int): The HTTP response code.\n
        servercode (int): The internal server error code.  Zero if none was provided by the server.\n
        servermessage (string): The internal server error message. Empty if none was provided by the server.\n
        serverdatajson (string): The internal server error data JSON. Empty if none was provided by the server.\n
        conversationid (string): The conversation ID generated by the client for use in the server request.
    """
    def __init__(self, httpcode, servercode, servermessage, serverdatajson, conversationid):
        ## (int) The HTTP response code
        self.httpcode = httpcode
        ## (int) The internal server error code or zero if none was provided by the server
        self.servercode = servercode
        ##(string) The internal server error message or empty if none was provided by the server
        self.servermessage = servermessage
        ##(string) The internal server error data JSON or empty if none was provided by the server
        self.serverdatajson = serverdatajson
        ##(string) The conversation ID generated by the client for use in the server request
        self.conversationid = conversationid
    
    @staticmethod
    def _marshalFromC(cServerResponseLP):
        if cServerResponseLP is None:
            return None
        cServerResponse = cServerResponseLP.contents
        return ServerResponse(cServerResponse.nHttpResponseCode,
                              cServerResponse.nServerErrorCode,
                              _private.CMarshalUtil.stringFromC(cServerResponse.pszServerErrorMessage),
                              _private.CMarshalUtil.stringFromC(cServerResponse.pszServerErrorDataJson),
                              _private.CMarshalUtil.stringFromC(cServerResponse.pszConversationId))
            
    @staticmethod
    def _marshalToC(serverException):
        if serverException is None:
            return None

        return CServerResponse( serverException.httpcode,
                                serverException.servercode,
                                _private.CMarshalUtil.stringToC(serverException.servermessage),
                                _private.CMarshalUtil.stringToC(serverException.serverdatajson),
                                _private.CMarshalUtil.stringToC(serverException.conversationid))

# Exception for server responses
class IonicServerException(IonicException, ServerResponse):
    """!
        Exception for Ionic server errors.

    Instance Variables:\n
        httpcode (int): The HTTP response code.\n
        servercode (int): The internal server error code.  Zero if none was provided by the server.\n
        servermessage (string): The internal server error message. Empty if none was provided by the server.\n
        serverdatajson (string): The internal server error data JSON. Empty if none was provided by the server.\n
        conversationid (string): The conversation ID generated by the client for use in the server request.
    """
    def __init__(self, message, code, httpcode, servercode, servermessage, serverdatajson, conversationid):
        IonicException.__init__(self, message, code)
        ServerResponse.__init__(self, httpcode, servercode, servermessage, serverdatajson, conversationid)
    
    def __str__(self):
        message = super(IonicServerException, self).__str__()
        if self.httpcode:
            message = message + " - HTTP code: " + str(self.httpcode)
        if self.servercode:
            message = message + " - Server error code: " + str(self.servercode)
        if self.servermessage:
            message = message + " - Server error message: " + str(self.servermessage)
        return message
    
    @staticmethod
    def _marshalFromC(message, code, cServerResponse):
        if cServerResponse is None:
            return None
        return IonicServerException(
            message,
            code,
            cServerResponse.nHttpResponseCode,
            cServerResponse.nServerErrorCode,
            _private.CMarshalUtil.stringFromC(cServerResponse.pszServerErrorMessage),
            _private.CMarshalUtil.stringFromC(cServerResponse.pszServerErrorDataJson),
            _private.CMarshalUtil.stringFromC(cServerResponse.pszConversationId))
        
